@@ -1,9 +1,15 @@
 #include "jsi.h"
-#include "jsvalue.h"
-#include "jsbuiltin.h"
 #include "utf.h"
 #include "regexp.h"
 
+static int js_doregexec(js_State *J, Reprog *prog, const char *string, Resub *sub, int eflags)
+{
+	int result = js_regexec(prog, string, sub, eflags);
+	if (result < 0)
+		js_error(J, "regexec failed");
+	return result;
+}
+
 static const char *checkstring(js_State *J, int idx)
 {
 	if (!js_iscoercible(J, idx))
@@ -13,32 +19,55 @@ static const char *checkstring(js_State *J, int idx)
 
 int js_runeat(js_State *J, const char *s, int i)
 {
-	Rune rune = 0;
-	while (i-- >= 0) {
+	Rune rune = EOF;
+	while (i >= 0) {
 		rune = *(unsigned char*)s;
 		if (rune < Runeself) {
 			if (rune == 0)
-				return 0;
+				return EOF;
 			++s;
-		} else
+			--i;
+		} else {
 			s += chartorune(&rune, s);
+			if (rune >= 0x10000)
+				i -= 2;
+			else
+				--i;
+		}
+	}
+	if (rune >= 0x10000) {
+		/* high surrogate */
+		if (i == -2)
+			return 0xd800 + ((rune - 0x10000) >> 10);
+		/* low surrogate */
+		else
+			return 0xdc00 + ((rune - 0x10000) & 0x3ff);
 	}
 	return rune;
 }
 
-const char *js_utfidxtoptr(const char *s, int i)
+int js_utflen(const char *s)
 {
+	int c;
+	int n;
 	Rune rune;
-	while (i-- > 0) {
-		rune = *(unsigned char*)s;
-		if (rune < Runeself) {
-			if (rune == 0)
-				return NULL;
-			++s;
-		} else
+
+	n = 0;
+	for(;;) {
+		c = *(unsigned char *)s;
+		if (c < Runeself) {
+			if (c == 0)
+				return n;
+			s++;
+			n++;
+		} else {
 			s += chartorune(&rune, s);
+			if (rune >= 0x10000)
+				n += 2;
+			else
+				n++;
+		}
 	}
-	return s;
 }
 
 int js_utfptrtoidx(const char *s, const char *p)
@@ -50,7 +79,10 @@ int js_utfptrtoidx(const char *s, const char *p)
 			++s;
 		else
 			s += chartorune(&rune, s);
-		++i;
+		if (rune >= 0x10000)
+			i += 2;
+		else
+			i += 1;
 	}
 	return i;
 }
@@ -69,14 +101,14 @@ static void Sp_toString(js_State *J)
 {
 	js_Object *self = js_toobject(J, 0);
 	if (self->type != JS_CSTRING) js_typeerror(J, "not a string");
-	js_pushliteral(J, self->u.s.string);
+	js_pushstring(J, self->u.s.string);
 }
 
 static void Sp_valueOf(js_State *J)
 {
 	js_Object *self = js_toobject(J, 0);
 	if (self->type != JS_CSTRING) js_typeerror(J, "not a string");
-	js_pushliteral(J, self->u.s.string);
+	js_pushstring(J, self->u.s.string);
 }
 
 static void Sp_charAt(js_State *J)
@@ -85,7 +117,7 @@ static void Sp_charAt(js_State *J)
 	const char *s = checkstring(J, 0);
 	int pos = js_tointeger(J, 1);
 	Rune rune = js_runeat(J, s, pos);
-	if (rune > 0) {
+	if (rune >= 0) {
 		buf[runetochar(buf, &rune)] = 0;
 		js_pushstring(J, buf);
 	} else {
@@ -98,7 +130,7 @@ static void Sp_charCodeAt(js_State *J)
 	const char *s = checkstring(J, 0);
 	int pos = js_tointeger(J, 1);
 	Rune rune = js_runeat(J, s, pos);
-	if (rune > 0)
+	if (rune >= 0)
 		js_pushnumber(J, rune);
 	else
 		js_pushnumber(J, NAN);
@@ -108,26 +140,31 @@ static void Sp_concat(js_State *J)
 {
 	int i, top = js_gettop(J);
 	int n;
-	char * volatile out;
+	char * volatile out = NULL;
 	const char *s;
 
 	if (top == 1)
 		return;
 
 	s = checkstring(J, 0);
-	n = strlen(s);
-	out = js_malloc(J, n + 1);
-	strcpy(out, s);
+	n = 1 + strlen(s);
 
 	if (js_try(J)) {
 		js_free(J, out);
 		js_throw(J);
 	}
 
+	if (n > JS_STRLIMIT)
+		js_rangeerror(J, "invalid string length");
+	out = js_malloc(J, n);
+	strcpy(out, s);
+
 	for (i = 1; i < top; ++i) {
 		s = js_tostring(J, i);
 		n += strlen(s);
-		out = js_realloc(J, out, n + 1);
+		if (n > JS_STRLIMIT)
+			js_rangeerror(J, "invalid string length");
+		out = js_realloc(J, out, n);
 		strcat(out, s);
 	}
 
@@ -159,7 +196,7 @@ static void Sp_lastIndexOf(js_State *J)
 {
 	const char *haystack = checkstring(J, 0);
 	const char *needle = js_tostring(J, 1);
-	int pos = js_isdefined(J, 2) ? js_tointeger(J, 2) : strlen(haystack);
+	int pos = js_isdefined(J, 2) ? js_tointeger(J, 2) : (int)strlen(haystack);
 	int len = strlen(needle);
 	int k = 0, last = -1;
 	Rune rune;
@@ -179,11 +216,67 @@ static void Sp_localeCompare(js_State *J)
 	js_pushnumber(J, strcmp(a, b));
 }
 
+static void Sp_substring_imp(js_State *J, const char *s, int a, int n)
+{
+	Rune head_rune = 0, tail_rune = 0;
+	const char *head, *tail;
+	char *p;
+	int i, k, head_len, tail_len;
+
+	/* find start of substring */
+	head = s;
+	for (i = 0; i < a; ++i) {
+		head += chartorune(&head_rune, head);
+		if (head_rune >= 0x10000)
+			++i;
+	}
+
+	/* find end of substring */
+	tail = head;
+	for (k = i - a; k < n; ++k) {
+		tail += chartorune(&tail_rune, tail);
+		if (tail_rune >= 0x10000)
+			++k;
+	}
+
+	/* no surrogate pair splits! */
+	if (i == a && k == n) {
+		js_pushlstring(J, head, tail - head);
+		return;
+	}
+
+	if (js_try(J)) {
+		js_free(J, p);
+		js_throw(J);
+	}
+
+	p = js_malloc(J, UTFmax + (tail - head));
+
+	/* substring starts with low surrogate (head is just after character) */
+	if (i > a) {
+		head_rune = 0xdc00 + ((head_rune - 0x10000) & 0x3ff);
+		head_len = runetochar(p, &head_rune);
+		memcpy(p + head_len, head, tail - head);
+		js_pushlstring(J, p, head_len + (tail - head));
+	}
+
+	/* substring ends with high surrogate (tail is just after character) */
+	if (k > n) {
+		tail -= runelen(tail_rune);
+		memcpy(p, head, tail - head);
+		tail_rune = 0xd800 + ((tail_rune - 0x10000) >> 10);
+		tail_len = runetochar(p + (tail - head), &tail_rune);
+		js_pushlstring(J, p, (tail - head) + tail_len);
+	}
+
+	js_endtry(J);
+	js_free(J, p);
+}
+
 static void Sp_slice(js_State *J)
 {
 	const char *str = checkstring(J, 0);
-	const char *ss, *ee;
-	int len = utflen(str);
+	int len = js_utflen(str);
 	int s = js_tointeger(J, 1);
 	int e = js_isdefined(J, 2) ? js_tointeger(J, 2) : len;
 
@@ -193,56 +286,48 @@ static void Sp_slice(js_State *J)
 	s = s < 0 ? 0 : s > len ? len : s;
 	e = e < 0 ? 0 : e > len ? len : e;
 
-	if (s < e) {
-		ss = js_utfidxtoptr(str, s);
-		ee = js_utfidxtoptr(ss, e - s);
-	} else {
-		ss = js_utfidxtoptr(str, e);
-		ee = js_utfidxtoptr(ss, s - e);
-	}
-
-	js_pushlstring(J, ss, ee - ss);
+	if (s < e)
+		Sp_substring_imp(J, str, s, e - s);
+	else
+		Sp_substring_imp(J, str, e, s - e);
 }
 
 static void Sp_substring(js_State *J)
 {
 	const char *str = checkstring(J, 0);
-	const char *ss, *ee;
-	int len = utflen(str);
+	int len = js_utflen(str);
 	int s = js_tointeger(J, 1);
 	int e = js_isdefined(J, 2) ? js_tointeger(J, 2) : len;
 
 	s = s < 0 ? 0 : s > len ? len : s;
 	e = e < 0 ? 0 : e > len ? len : e;
 
-	if (s < e) {
-		ss = js_utfidxtoptr(str, s);
-		ee = js_utfidxtoptr(ss, e - s);
-	} else {
-		ss = js_utfidxtoptr(str, e);
-		ee = js_utfidxtoptr(ss, s - e);
-	}
-
-	js_pushlstring(J, ss, ee - ss);
+	if (s < e)
+		Sp_substring_imp(J, str, s, e - s);
+	else
+		Sp_substring_imp(J, str, e, s - e);
 }
 
 static void Sp_toLowerCase(js_State *J)
 {
-	const char *src = checkstring(J, 0);
-	char *dst = js_malloc(J, UTFmax * strlen(src) + 1);
-	const char *s = src;
-	char *d = dst;
+	const char *s = checkstring(J, 0);
+	char * volatile dst = NULL;
+	char *d;
 	Rune rune;
+
+	if (js_try(J)) {
+		js_free(J, dst);
+		js_throw(J);
+	}
+
+	d = dst = js_malloc(J, UTFmax * strlen(s) + 1);
 	while (*s) {
 		s += chartorune(&rune, s);
 		rune = tolowerrune(rune);
 		d += runetochar(d, &rune);
 	}
 	*d = 0;
-	if (js_try(J)) {
-		js_free(J, dst);
-		js_throw(J);
-	}
+
 	js_pushstring(J, dst);
 	js_endtry(J);
 	js_free(J, dst);
@@ -250,21 +335,24 @@ static void Sp_toLowerCase(js_State *J)
 
 static void Sp_toUpperCase(js_State *J)
 {
-	const char *src = checkstring(J, 0);
-	char *dst = js_malloc(J, UTFmax * strlen(src) + 1);
-	const char *s = src;
-	char *d = dst;
+	const char *s = checkstring(J, 0);
+	char * volatile dst = NULL;
+	char *d;
 	Rune rune;
+
+	if (js_try(J)) {
+		js_free(J, dst);
+		js_throw(J);
+	}
+
+	d = dst = js_malloc(J, UTFmax * strlen(s) + 1);
 	while (*s) {
 		s += chartorune(&rune, s);
 		rune = toupperrune(rune);
 		d += runetochar(d, &rune);
 	}
 	*d = 0;
-	if (js_try(J)) {
-		js_free(J, dst);
-		js_throw(J);
-	}
+
 	js_pushstring(J, dst);
 	js_endtry(J);
 	js_free(J, dst);
@@ -291,23 +379,24 @@ static void Sp_trim(js_State *J)
 static void S_fromCharCode(js_State *J)
 {
 	int i, top = js_gettop(J);
+	char * volatile s = NULL;
+	char *p;
 	Rune c;
-	char *s, *p;
-
-	s = p = js_malloc(J, (top-1) * UTFmax + 1);
 
 	if (js_try(J)) {
 		js_free(J, s);
 		js_throw(J);
 	}
 
+	s = p = js_malloc(J, (top-1) * UTFmax + 1);
+
 	for (i = 1; i < top; ++i) {
-		c = js_touint16(J, i);
+		c = js_touint32(J, i);
 		p += runetochar(p, &c);
 	}
 	*p = 0;
-	js_pushstring(J, s);
 
+	js_pushstring(J, s);
 	js_endtry(J);
 	js_free(J, s);
 }
@@ -343,7 +432,7 @@ static void Sp_match(js_State *J)
 	a = text;
 	e = text + strlen(text);
 	while (a <= e) {
-		if (js_regexec(re->prog, a, &m, a > text ? REG_NOTBOL : 0))
+		if (js_doregexec(J, re->prog, a, &m, a > text ? REG_NOTBOL : 0))
 			break;
 
 		b = m.sub[0].sp;
@@ -356,6 +445,11 @@ static void Sp_match(js_State *J)
 		if (c - b == 0)
 			++a;
 	}
+
+	if (len == 0) {
+		js_pop(J, 1);
+		js_pushnull(J);
+	}
 }
 
 static void Sp_search(js_State *J)
@@ -375,7 +469,7 @@ static void Sp_search(js_State *J)
 
 	re = js_toregexp(J, -1);
 
-	if (!js_regexec(re->prog, text, &m, 0))
+	if (!js_doregexec(J, re->prog, text, &m, 0))
 		js_pushnumber(J, js_utfptrtoidx(text, m.sub[0].sp));
 	else
 		js_pushnumber(J, -1);
@@ -392,7 +486,7 @@ static void Sp_replace_regexp(js_State *J)
 	source = checkstring(J, 0);
 	re = js_toregexp(J, 1);
 
-	if (js_regexec(re->prog, source, &m, 0)) {
+	if (js_doregexec(J, re->prog, source, &m, 0)) {
 		js_copy(J, 0);
 		return;
 	}
@@ -405,7 +499,7 @@ loop:
 
 	if (js_iscallable(J, 2)) {
 		js_copy(J, 2);
-		js_pushundefinedthis(J);
+		js_pushundefined(J);
 		for (x = 0; m.sub[x].sp; ++x) /* arg 0..x: substring and subexps that matched */
 			js_pushlstring(J, m.sub[x].sp, m.sub[x].ep - m.sub[x].sp);
 		js_pushnumber(J, s - source); /* arg x+2: offset within search string */
@@ -421,7 +515,8 @@ loop:
 		while (*r) {
 			if (*r == '$') {
 				switch (*(++r)) {
-				case 0: --r; /* end of string; back up and fall through */
+				case 0: --r; /* end of string; back up */
+				/* fallthrough */
 				case '$': js_putc(J, &sb, '$'); break;
 				case '`': js_putm(J, &sb, source, s); break;
 				case '\'': js_puts(J, &sb, s + n); break;
@@ -465,7 +560,7 @@ loop:
 			else
 				goto end;
 		}
-		if (!js_regexec(re->prog, source, &m, REG_NOTBOL))
+		if (!js_doregexec(J, re->prog, source, &m, REG_NOTBOL))
 			goto loop;
 	}
 
@@ -500,7 +595,7 @@ static void Sp_replace_string(js_State *J)
 
 	if (js_iscallable(J, 2)) {
 		js_copy(J, 2);
-		js_pushundefinedthis(J);
+		js_pushundefined(J);
 		js_pushlstring(J, s, n); /* arg 1: substring that matched */
 		js_pushnumber(J, s - source); /* arg 2: offset within search string */
 		js_copy(J, 0); /* arg 3: search string */
@@ -517,7 +612,8 @@ static void Sp_replace_string(js_State *J)
 		while (*r) {
 			if (*r == '$') {
 				switch (*(++r)) {
-				case 0: --r; /* end of string; back up and fall through */
+				case 0: --r; /* end of string; back up */
+				/* fallthrough */
 				case '$': js_putc(J, &sb, '$'); break;
 				case '&': js_putm(J, &sb, s, s + n); break;
 				case '`': js_putm(J, &sb, source, s); break;
@@ -565,12 +661,14 @@ static void Sp_split_regexp(js_State *J)
 	js_newarray(J);
 	len = 0;
 
+	if (limit == 0)
+		return;
+
 	e = text + strlen(text);
 
 	/* splitting the empty string */
 	if (e == text) {
-		if (js_regexec(re->prog, text, &m, 0)) {
-			if (len == limit) return;
+		if (js_doregexec(J, re->prog, text, &m, 0)) {
 			js_pushliteral(J, "");
 			js_setindex(J, -2, 0);
 		}
@@ -579,14 +677,14 @@ static void Sp_split_regexp(js_State *J)
 
 	p = a = text;
 	while (a < e) {
-		if (js_regexec(re->prog, a, &m, a > text ? REG_NOTBOL : 0))
+		if (js_doregexec(J, re->prog, a, &m, a > text ? REG_NOTBOL : 0))
 			break; /* no match */
 
 		b = m.sub[0].sp;
 		c = m.sub[0].ep;
 
 		/* empty string at end of last match */
-		if (b == p) {
+		if (b == c && b == p) {
 			++a;
 			continue;
 		}
@@ -618,6 +716,9 @@ static void Sp_split_string(js_State *J)
 
 	js_newarray(J);
 
+	if (limit == 0)
+		return;
+
 	n = strlen(sep);
 
 	/* empty string */
@@ -650,7 +751,7 @@ static void Sp_split(js_State *J)
 {
 	if (js_isundefined(J, 1)) {
 		js_newarray(J);
-		js_copy(J, 0);
+		js_pushstring(J, js_tostring(J, 0));
 		js_setindex(J, -2, 0);
 	} else if (js_isregexp(J, 1)) {
 		Sp_split_regexp(J);
@@ -661,7 +762,8 @@ static void Sp_split(js_State *J)
 
 void jsB_initstring(js_State *J)
 {
-	J->String_prototype->u.s.string = "";
+	J->String_prototype->u.s.shrstr[0] = 0;
+	J->String_prototype->u.s.string = J->String_prototype->u.s.shrstr;
 	J->String_prototype->u.s.length = 0;
 
 	js_pushobject(J, J->String_prototype);
